# Convert plain text to html by adding <p> and </p>
# Also handle \(shiny)Rule and \(shiny)Small\Med\BigSkip
# Also handle \begin/end{enumerate/itemize} and \item
# Also handle tooltips
# Actually this function mostly distiguishes verbatim, blanks, and other text (which is sent for processing)
htmlify = function(tex, lineNumOffset) {
  # handle terminal "\\"
  tex = gsub("[\\]{2,4}[[:blank:]]*$", "<p></p>", tex)

  # Handle verbatim environment (perhaps not well!)  Cannot have >1 verbatim per line and must be alone (seems OK)
  verbB = regexpr("[\\]begin[[:blank:]]*[{]verbatim[}]", tex)
  verbE = regexpr("[\\]end[[:blank:]]*[{]verbatim[}]", tex)
  SelB = which(verbB != -1)
  SelE = which(verbE != -1)
  if (length(SelB) != length(SelE)) {
    stop("@", lineNumOffset, "-",lineNumOffset+length(tex)-1, " unmatched begin/end{verbatim}s")
  }
  if (length(SelB) == 0) {
    lineType = rep("t", length(tex)) # "t" for (plain) text
  } else {
    nv = length(SelB)
    temp = grep("^[[:blank:]]*[\\]begin[[:blank:]]*[{]verbatim[}][[:blank:]]*$", tex[SelB], invert=TRUE)
    if (length(temp) > 0) {
      stop("@", SelB[temp[1]]-1+lineNumOffset, " \\begin{verbatim} must be alone on its line")
    }
    temp = grep("^[[:blank:]]*[\\]end[[:blank:]]*[{]verbatim[}][[:blank:]]*$", tex[SelE], invert=TRUE)
    if (length(temp) > 0) {
      stop("@", SelE[temp[1]]-1+lineNumOffset, " \\end{verbatim} must be alone on its line")
    }
    temp = as.vector(t(cbind(SelB,SelE)))
    temp2 = diff(temp)
    if (any(temp2 < 0)) {
      stop("@", temp[which(temp2<0)[1]+1]-1+lineNumOffset, " misordered {verbatim} environments")
    }
    # special handling (?waste of time) for adjacent verbatims
    if (nv > 1) {
      for (i in nv:2) {
        if (SelB[i] == SelE[i-1]+1)
          tex = c(tex[1:SelE[i-1]], '', tex[SelB[i]:length(tex)])
      }
    }
    lineType = rep("t", length(tex)) # "t" for (plain) text
    for (i in 1:nv) lineType[SelB[i]:SelE[i]] = "v"
  }
  
  #
  # parse sections (blank vs. verbatim vs. text with mathJax handling)
  isBlank = regexpr("^[[:blank:]]*$", tex) != -1
  lineType[isBlank & lineType!="v"] = "b"
  rleType = rle(lineType)
  blockType = rleType$values
  m = length(rleType$lengths)
  if (m==1 && rleType$values==TRUE) return(NULL)
  typeEnd = cumsum(rleType$lengths)
  if (m==1) {
    typeBegin = 1
  } else {
    typeBegin = c(1, typeEnd[1:(m-1)]+1)
  }
    
  rtn = NULL
  for (block in 1:m) {
    if (blockType[block] == "v") { # verbatim
      rtn = c(rtn, "<pre>", tex[(typeBegin[block]+1):(typeEnd[block]-1)], "</pre>")
    } else if (blockType[block] == "b") { # blank
      next
    } else if (blockType[block] == "t") { # text
      texRange = paste0("@", typeBegin[block]-1+lineNumOffset ,"-", typeEnd[block]-1+lineNumOffset)
      rtn = c(rtn, "<p>", 
              processText(paste(tex[typeBegin[block]:typeEnd[block]], collapse=" "), texRange),
              "</p>")
    } else {
      cat("lineType ", blockType[block], " not handled\n")
      assure(FALSE, 99231)
    }
  }

  return(rtn)
} # end htmlify()
